# -*- coding:utf-8 -*-
from collections import OrderedDict
import sqlalchemy.types as t
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.sql.visitors import VisitableType
from sqlalchemy.orm.base import ONETOMANY
from sqlalchemy.sql.type_api import TypeEngine


class InvalidStatus(Exception):
    pass

"""
http://json-schema.org/latest/json-schema-core.html#anchor8
3.5.  JSON Schema primitive types

JSON Schema defines seven primitive types for JSON values:

    array
        A JSON array.
    boolean
        A JSON boolean.
    integer
        A JSON number without a fraction or exponent part.
    number
        Any JSON number. Number includes integer.
    null
        The JSON null value.
    object
        A JSON object.
    string
        A JSON string.
"""

#  tentative
default_column_to_schema = {
    t.String: "string",
    t.Text: "string",
    t.Integer: "integer",
    t.SmallInteger: "integer",
    t.BigInteger: "string",  # xxx
    t.Numeric: "integer",
    t.Float: "number",
    t.DateTime: "string",
    t.Date: "string",
    t.Time: "string",  # xxx
    t.LargeBinary: "xxx",
    t.Binary: "xxx",
    t.Boolean: "boolean",
    t.Unicode: "string",
    t.Concatenable: "xxx",
    t.UnicodeText: "string",
    t.Interval: "xxx",
    t.Enum: "string",
}


# restriction
def string_max_length(column, sub):
    sub["maxLength"] = column.type.length


def enum_one_of(column, sub):
    sub["enum"] = list(column.type.enums)


def datetime_format(column, sub):
    sub["format"] = "date-time"


def date_format(column, sub):
    sub["format"] = "date"


def time_format(column, sub):
    sub["format"] = "time"


default_restriction_dict = {
    t.String: string_max_length,
    t.Enum: enum_one_of,
    t.DateTime: datetime_format,
    t.Date: date_format,
    t.Time: time_format
}


class Classifier(object):
    def __init__(self, mapping=default_column_to_schema):
        self.mapping = mapping

    def __getitem__(self, k):
        cls = k.__class__
        v = self.mapping.get(cls)
        if v is not None:
            return cls, v
        # inheritance
        for type_ in self.mapping:
            if issubclass(cls, type_):
                return type_, self.mapping[type_]
        raise InvalidStatus("notfound: {k}".format(k=k))

DefaultClassfier = Classifier(default_column_to_schema)
Empty = ()


class BaseModelWalker(object):
    def __init__(self, model, includes=None, excludes=None, history=None):
        self.mapper = inspect(model).mapper
        self.includes = includes
        self.excludes = excludes
        self.history = history or []
        if includes and excludes:
            if set(includes).intersection(excludes):
                raise InvalidStatus("Conflict includes={}, exclude={}".format(includes, excludes))

# mapper.column_attrs and mapper.attrs is not ordered. define our custom iterate function `iterate'


class SingleModelWalker(BaseModelWalker):
    def iterate(self):
        for c in self.mapper.local_table.columns:
            yield self.mapper._props[c.name]  # danger!! not immutable

    def walk(self):
        for prop in self.iterate():
            if self.includes is None or prop.key in self.includes:
                if self.excludes is None or prop.key not in self.excludes:
                    yield prop


class OneModelOnlyWalker(BaseModelWalker):
    def iterate(self):
        for c in self.mapper.local_table.columns:
            yield self.mapper._props[c.name]  # danger!! not immutable

    def walk(self):
        for prop in self.iterate():
            if self.includes is None or prop.key in self.includes:
                if self.excludes is None or prop.key not in self.excludes:
                    if not any(c.foreign_keys for c in getattr(prop, "columns", Empty)):
                        yield prop


class AlsoChildrenWalker(BaseModelWalker):
    def iterate(self):
        # self.mapper.attrs
        for c in self.mapper.local_table.columns:
            yield self.mapper._props[c.name]  # danger!! not immutable
        for prop in self.mapper.relationships:
            yield prop

    def walk(self):
        for prop in self.iterate():
            if isinstance(prop, (ColumnProperty, RelationshipProperty)):
                if self.includes is None or prop.key in self.includes:
                    if self.excludes is None or prop.key not in self.excludes:
                        if prop not in self.history:
                            if not any(c.foreign_keys for c in getattr(prop, "columns", Empty)):
                                yield prop

pop_marker = object()


def get_children(name, params, splitter=".", default=None):  # todo: rename
    prefix = name + splitter
    if hasattr(params, "items"):
        return {k.split(splitter, 1)[0]: v for k, v in params.items() if k.startswith(prefix)}
    elif isinstance(params, (list, tuple)):
        return [e.split(splitter, 1)[0] for e in params if e.startswith(prefix)]
    else:
        return default


class CollectionForOverrides(object):
    def __init__(self, params, pop_marker=pop_marker):
        self.params = params or {}
        self.not_used_keys = set(params.keys())
        self.pop_marker = pop_marker

    def __contains__(self, k):
        return k in self.params

    def overrides(self, basedict):
        for k, v in self.params.items():
            if v == self.pop_marker:
                basedict.pop(k)  # xxx: KeyError?
            else:
                basedict[k] = v
            self.not_used_keys.remove(k)  # xxx: KeyError?


class ChildFactory(object):
    def __init__(self, splitter="."):
        self.splitter = splitter

    def default_excludes(self, prop):
        return [prop.back_populates, prop.backref]

    def child_overrides(self, prop, overrides):
        name = prop.key
        children = get_children(name, overrides.params, splitter=self.splitter)
        return overrides.__class__(children, pop_marker=overrides.pop_marker)

    def child_walker(self, prop, walker, history=None):
        name = prop.key
        excludes = get_children(name, walker.includes, splitter=self.splitter, default=[])
        excludes.extend(self.default_excludes(prop))
        includes = get_children(name, walker.includes, splitter=self.splitter)

        return walker.__class__(prop.mapper, includes=includes, excludes=excludes, history=history)

    def child_schema(self, prop, schema_factory, walker, overrides, depth, history):
        subschema = schema_factory._build_properties(walker, overrides, depth=(depth and depth - 1), history=history)
        if prop.direction == ONETOMANY:
            return {"type": "array", "items": subschema}
        else:
            return subschema


class SchemaFactory(object):
    def __init__(self, walker,
                 classifier=DefaultClassfier,
                 restriction_dict=default_restriction_dict,
                 container_factory=OrderedDict,
                 child_factory=ChildFactory(".")):
        self.container_factory = container_factory
        self.classifier = classifier
        self.walker = walker  # class
        self.restriction_dict = restriction_dict
        self.child_factory = child_factory

    def __call__(self, model, includes=None, excludes=None, overrides=None, depth=None):
        walker = self.walker(model, includes=includes, excludes=excludes)
        overrides = CollectionForOverrides(overrides or {})

        schema = {
            "title": model.__name__,
            "type": "object",
            "properties": self._build_properties(walker, overrides=overrides, depth=depth)
        }

        if overrides.not_used_keys:
            raise InvalidStatus("invalid overrides: {}".format(overrides.not_used_keys))

        if model.__doc__:
            schema["description"] = model.__doc__

        required = self._detect_required(walker)

        if required:
            schema["required"] = required
        return schema

    def _add_restriction_if_found(self, D, column, itype):
        for tcls in itype.__mro__:
            if tcls is TypeEngine:
                break
            fn = self.restriction_dict.get(tcls)
            if fn is not None:
                fn(column, D)

    def _build_properties(self, walker, overrides, depth=None, history=None):
        if depth is not None and depth <= 0:
            return self.container_factory()

        D = self.container_factory()
        if history is None:
            history = []

        for prop in walker.walk():
            if hasattr(prop, "mapper"):     # RelationshipProperty
                history.append(prop)
                subwalker = self.child_factory.child_walker(prop, walker, history=history)
                suboverrides = self.child_factory.child_overrides(prop, overrides)
                D[prop.key] = self.child_factory.child_schema(prop, self, subwalker, suboverrides, depth=depth, history=history)
                history.pop()
            elif hasattr(prop, "columns"):  # ColumnProperty
                for c in prop.columns:
                    sub = {}
                    if type(c.type) != VisitableType:
                        itype, sub["type"] = self.classifier[c.type]

                        self._add_restriction_if_found(sub, c, itype)

                        if c.doc:
                            sub["description"] = c.doc

                        if c.name in overrides:
                            overrides.overrides(sub)

                        D[c.name] = sub
                    else:
                        raise NotImplemented
                D[prop.key] = sub
        return D

    def _detect_required(self, walker):
        return [prop.key for prop in walker.walk() if any(not c.nullable for c in getattr(prop, "columns", Empty))]



# -*- coding:utf-8 -*-
import sys
import pkg_resources
import json
import argparse


from . import (
    SchemaFactory,
    AlsoChildrenWalker,
    OneModelOnlyWalker,
    SingleModelWalker
)


def import_symbol(symbol):  # todo cache
    return pkg_resources.EntryPoint.parse("x=%s" % symbol).load(False)


def err(x):
    sys.stderr.write(x)
    sys.stderr.write("\n")
    sys.stderr.flush()


def detect_walker(x):
    if x == "alsochildren":
        return AlsoChildrenWalker
    elif x == "onlyone":
        return OneModelOnlyWalker
    elif x == "single":
        return SingleModelWalker
    else:
        raise Exception(x)


def run(args):
    model = import_symbol(args.model)
    walker = detect_walker(args.walker)
    make_schema = SchemaFactory(walker)
    schema = make_schema(model)
    print(json.dumps(schema, indent=2, ensure_ascii=False))


def main(sys_args=sys.argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("program")
    parser.add_argument("model")
    parser.add_argument("--walker", choices=["onlyone", "single", "alsochildren"], default="alsochildren")
    args = parser.parse_args(sys_args)
    return run(args)



# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)
from .compat import text_
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.relationships import RelationshipProperty
from functools import partial
import isodate
from .custom.format import (
    parse_time,  # more strict than isodate
    parse_date   # more strict
)
from collections import OrderedDict
from jsonschema import validate
from . import InvalidStatus
import pytz


class ConvertionError(Exception):
    def __init__(self, name, message):
        self.name = name
        self.message = message


def datetime_rfc3339(ob):
    if ob.tzinfo:
        return ob.isoformat()
    return pytz.utc.localize(ob).isoformat()


def isoformat(ob):
    return ob.isoformat() + "Z"  # xxx


def isoformat0(ob):
    return ob.isoformat()


def raise_error(ob):
    raise Exception("convert failure. unknown format xxx of {}".format(ob))


def maybe_wrap(fn, default=None):
    def wrapper(ob):
        if ob is None:
            return default
        return fn(ob)
    return wrapper

# todo: look at required or not
jsonify_dict = {('string', None): maybe_wrap(text_),
                ('string', 'time',): maybe_wrap(isoformat),
                ('number', None): maybe_wrap(float),
                ('integer', None): maybe_wrap(int),
                ('boolean', None): maybe_wrap(bool),
                ('string', 'date-time'): maybe_wrap(datetime_rfc3339),
                ('string', 'date'): maybe_wrap(isoformat0),
                ('xxx', None): raise_error}


normalize_dict = {('string', None): maybe_wrap(text_),
                  ('string', 'time',): maybe_wrap(parse_time),
                  ('number', None): maybe_wrap(float),
                  ('integer', None): maybe_wrap(int),
                  ('boolean', None): maybe_wrap(bool),
                  ('string', 'date-time'): maybe_wrap(isodate.parse_datetime),
                  ('string', 'date'): maybe_wrap(parse_date),
                  ('xxx', None): raise_error}

prepare_dict = {'string': maybe_wrap(text_),
                'number': maybe_wrap(float),
                'integer': maybe_wrap(int),
                'boolean': maybe_wrap(bool)}


def jsonify_of(ob, name, type_, registry=jsonify_dict):
    try:
        convert_fn = registry[type_]
    except KeyError:
        raise ConvertionError(name, "convert {} failure. unknown format {} of {}".format(name, type_, ob))
    return convert_fn(getattr(ob, name, None))

marker = object()


def normalize_of(ob, name, type_, registry=normalize_dict):
    try:
        convert_fn = registry[type_]
    except KeyError:
        raise ConvertionError(name, "convert {} failure. unknown format {} of {}".format(name, type_, ob))
    try:
        val = ob.get(name, marker)
        if val is marker:
            return val
        return convert_fn(val)
    except ValueError as e:
        raise ConvertionError(name, e.args[0])


def prepare_of(ob, name, type_, registry=prepare_dict):
    val = ob.get(name, marker)
    if val is marker:
        return val
    try:
        convert_fn = registry[type_[0]]
        return convert_fn(val)
    except KeyError:
        return val
    except ValueError as e:
        raise ConvertionError(name, e.args[0])


def attribute_of(ob, name, type_):
    return getattr(ob, name)


def dictify(ob, schema, convert=attribute_of, getter=getattr):
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def jsonify(ob, schema, convert=jsonify_of, getter=getattr, registry=jsonify_dict):
    convert = partial(convert, registry=registry)
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def normalize(ob, schema, convert=normalize_of, getter=dict.get, registry=normalize_dict):
    convert = partial(convert, registry=registry)
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def prepare(ob, schema, convert=prepare_of, getter=dict.get, registry=prepare_dict):
    convert = partial(convert, registry=registry)
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def dictify_properties(ob, properties, convert, getter, marker=marker):
    if ob is None:
        return None
    D = {}
    for k, v in properties.items():
        val = _dictify(ob, k, v, convert, getter)
        if val is not marker:
            D[k] = val
    return D


def _dictify(ob, name, schema, convert, getter):
    type_ = schema.get("type")
    if type_ == "array":
        return [dictify_properties(e, schema["items"], convert, getter) for e in getter(ob, name, [])]
    elif type_ is None:
        return dictify_properties(getter(ob, name), schema, convert, getter)
    elif type_ == "object":
        return dictify_properties(getter(ob, name), schema["properties"], convert, getter)
    else:
        return convert(ob, name, (type_, schema.get("format")))


class ModelLookup(object):
    def __init__(self, module):
        self.module = module
        self.name_stack = []
        self.inspect_stack = []

    def __call__(self, name):
        if not self.name_stack:
            self.name_stack.append(name)
            model = getattr(self.module, name)
            self.inspect_stack.append(inspect(model))
            return model
        else:
            self.name_stack.append(name)
            prop = self.inspect_stack[-1].get_property(name)
            assert isinstance(prop, RelationshipProperty)
            mapper = prop.mapper
            model = mapper.class_
            self.inspect_stack.append(mapper)
            return model

    def pop(self):
        name = self.name_stack.pop()
        return name, self.inspect_stack.pop()


class ComposedModule(object):
    def __init__(self, *modules):
        self.modules = set(modules)

    def __getattr__(self, k):
        for m in self.modules:
            if hasattr(m, k):
                return getattr(m, k)


# objectify
def _objectify_subobject(params, name, schema, modellookup):
    sub_model = modellookup(name)
    sub_params = objectify_propperties(params, schema, modellookup)
    sub = sub_model(**sub_params)
    modellookup.pop()
    return sub


def objectify(params, schema, modellookup, strict=True):
    model_class = modellookup(schema["title"])
    params = objectify_propperties(params, schema["properties"], modellookup)
    result = model_class(**params)
    modellookup.pop()
    if strict:
        for k in schema["required"]:
            if getattr(result, k) is None:
                raise InvalidStatus("{}.{} is None. this is required.".format(model_class, k))
    assert modellookup.name_stack == []
    return result


def objectify_propperties(params, properties, modellookup):
    D = {}
    for k, schema in properties.items():
        D[k] = _objectify(params, k, schema, modellookup)
    return D


def _objectify(params, name, schema, modellookup):
    type_ = schema.get("type")
    if params is None:
        return [] if type_ == "array" else None  # xxx

    if type_ == "array":
        sub_schema = schema["items"]
        return [_objectify_subobject(e, name, sub_schema, modellookup) for e in params.get(name, [])]
    elif name not in params:
        return None
    elif type_ is None:  # object
        sub_params = params.get(name)
        if sub_params is None:
            return None
        return _objectify_subobject(params[name], name, schema, modellookup)
    else:
        return params.get(name)


# apply_changes
def _apply_changes_subobject(parent, params, name, schema, modellookup):
    if params is None:
        return None
    sub = getattr(parent, name, None)
    if sub is None:
        return _objectify_subobject(params, name, schema, modellookup)
    else:
        sub_model = modellookup(name)
        assert sub.__class__ == sub_model
        sub = apply_changes_propperties(sub, params, schema, modellookup)
        modellookup.pop()
        return sub


def apply_changes(ob, params, schema, modellookup):
    model_class = modellookup(schema["title"])
    params = apply_changes_propperties(ob, params, schema["properties"], modellookup)
    assert model_class == ob.__class__
    modellookup.pop()
    assert modellookup.name_stack == []
    return ob


def apply_changes_propperties(ob, params, properties, modellookup):
    for k, schema in properties.items():
        setattr(ob, k, _apply_changes(ob, params, k, schema, modellookup))
    return ob


def _get_primary_keys_from_object(ob):
    return tuple(sorted(col.name for col in ob.__mapper__.primary_key))


def _get_primary_keys_from_params(sub_params, primary_keys):
    return tuple(sorted(sub_params.get(k) for k in primary_keys))


def subobject_iterate(ob, params, name):
    cache = OrderedDict()
    used_children = set()

    primary_keys = None

    for sub in getattr(ob, name):
        if primary_keys is None:
            primary_keys = _get_primary_keys_from_object(ob)
        cache[primary_keys] = sub

    for sub_params in params.get(name, []):
        keys = _get_primary_keys_from_params(sub_params, primary_keys)
        if keys in cache:
            sub = cache[keys]
            yield "update", sub, sub_params
            used_children.add(sub)
        else:
            yield "create", None, sub_params

    for sub in getattr(ob, name):
        if sub not in used_children:
            yield "delete", sub, None


def _apply_changes(ob, params, name, schema, modellookup):
    type_ = schema.get("type")
    if params is None:
        return [] if type_ == "array" else None  # xxx

    if type_ == "array":
        sub_schema = schema["items"]
        access = getattr(ob, name)
        for ac, sub, sub_params in list(subobject_iterate(ob, params, name)):
            if ac == "create":
                access.append(_objectify_subobject(sub_params, name, sub_schema, modellookup))
            elif ac == "update":
                for k, v in sub_params.items():  # xxx:
                    setattr(sub, k, v)
            elif ac == "delete":
                access.remove(sub)
        return getattr(ob, name)
    elif name not in params:
        return None
    elif type_ is None:  # object
        sub_params = params.get(name)
        if sub_params is None:
            return None
        return _apply_changes_subobject(ob, params[name], name, schema, modellookup)
    else:
        return params.get(name)


class ErrorFound(Exception):  # xxx:
    def __init__(self, errors):
        self.errors = errors


def raise_error(data, e):
    raise e


def validate_all(data, validator, treat_error=raise_error):
    errors = []
    for e in validator.iter_errors(data):
        errors.append(e)
    if errors:
        return treat_error(data, ErrorFound(errors))
    return data



# -*- coding:utf-8 -*-
import re
from jsonschema._format import _checks_drafts, FormatChecker, _draft_checkers
import calendar
from datetime import date, time
from ..compat import string_types

"""
this is custom format
"""
time_rx = re.compile("(\d{2}):(\d{2}):(\d{2})(\.\d+)?(Z|([+\-])(\d{2}):(\d{2}))?")
date_rx = re.compile("(\d{4})\-(\d{2})\-(\d{2})")


def parse_date(date_string):
    m = date_rx.match(date_string)
    if m is None:
        return None

    groups = m.groups()

    year, month, day = [int(x) for x in groups[:3]]
    return date(year, month, day)


def validate_date(date_string):
    m = date_rx.match(date_string)
    if m is None:
        return False

    groups = m.groups()

    year, month, day = [int(x) for x in groups[:3]]

    if not 1 <= year <= 9999:
        # Have to reject this, unfortunately (despite it being OK by rfc3339):
        # calendar.timegm/calendar.monthrange can't cope (since datetime can't)
        return False

    if not 1 <= month <= 12:
        return False

    (_, max_day) = calendar.monthrange(year, month)
    if not 1 <= day <= max_day:
        return False

    # all OK
    return True


def parse_time(time_string):
    m = time_rx.match(time_string)
    if m is None:
        return None

    groups = m.groups()

    hour, minute, second = [int(x) for x in groups[:3]]
    if groups[4] is not None and groups[4] != "Z":
        return time(hour, minute, second, int(groups(3)))
    return time(hour, minute, second)


def validate_time(time_string):
    m = time_rx.match(time_string)
    if m is None:
        return False

    groups = m.groups()

    hour, minute, second = [int(x) for x in groups[:3]]
    if not (0 <= hour <= 23 and 0 <= minute <= 59 and 0 <= second <= 59):
        # forbid leap seconds :-(. See README
        return False

    if groups[4] is not None and groups[4] != "Z":
        (offset_sign, offset_hours, offset_mins) = groups[5:]
        if not (0 <= int(offset_hours) <= 23 and 0 <= int(offset_mins) <= 59):
            return False

    # all OK
    return True


@_checks_drafts("date", raises=ValueError)
def is_date(instance):
    if not isinstance(instance, string_types):
        return True
    return validate_date(instance)


@_checks_drafts("time", raises=ValueError)
def is_time(instance):
    if not isinstance(instance, string_types):
        return True
    return validate_time(instance)


draft3_format_checker = FormatChecker(_draft_checkers["draft3"])
draft4_format_checker = FormatChecker(_draft_checkers["draft4"])



# -*- coding:utf-8 -*-
from functools import partial
from .dictify import (
    objectify,
    jsonify,
    dictify,
    normalize,
    prepare,
    validate_all,
    ModelLookup,
    jsonify_dict,
    normalize_dict,
    prepare_dict,
    raise_error
)
from jsonschema import (
    validate,
    FormatChecker
)
from jsonschema.validators import (
    Draft3Validator,
    Draft4Validator
)
from . import (
    default_restriction_dict,
    default_column_to_schema
)


class DefaultRegistry:
    jsonify = jsonify_dict
    normalize = normalize_dict
    prepare = prepare_dict
    restriction = default_restriction_dict
    column_to_schema = default_column_to_schema


class Mapping(object):
    def __init__(self, validator, model, modellookup, registry=DefaultRegistry, treat_error=raise_error):
        self.validator = validator
        self.format_checker = validator.format_checker
        self.schema = validator.schema
        self.model = model
        self.modellookup = modellookup
        self.registry = registry
        self.treat_error = treat_error

    def jsondict_from_object(self, ob):
        return jsonify(ob, self.schema, registry=self.registry.jsonify)

    def jsondict_from_string_only_dict(self, string_only_dict):
        return prepare(string_only_dict, self.schema, registry=self.registry.prepare)

    def dict_from_jsondict(self, jsondict):
        return normalize(jsondict, self.schema, registry=self.registry.normalize)

    def dict_from_object(self, ob):
        return dictify(ob, self.schema)

    def object_from_dict(self, params, strict=True):
        return objectify(params, self.schema, self.modellookup, strict=strict)

    def validate_jsondict(self, jsondict):
        return validate(jsondict, self.schema, format_checker=self.format_checker)

    def validate_all_jsondict(self, jsondict):
        return validate_all(jsondict, self.validator, self.treat_error)


class MappingFactory(object):
    _Mapping = Mapping
    _ModelLookup = ModelLookup

    def __init__(self, validator_class, schema_factory, module, resolver=None, format_checker=None):
        self.schema_factory = schema_factory
        self.validator_class = validator_class
        self.resolver = resolver
        self.format_checker = format_checker or FormatChecker()
        self.module = module

    def __call__(self, model, includes=None, excludes=None, depth=None):
        schema = self.schema_factory(model, includes=includes, excludes=excludes, depth=depth)
        validator = self.validator_class(schema, resolver=self.resolver, format_checker=self.format_checker)
        modellookup = self._ModelLookup(self.module)
        mapping = self._Mapping(validator, model, modellookup)
        return mapping

Draft3MappingFactory = partial(MappingFactory, Draft3Validator)
Draft4MappingFactory = partial(MappingFactory, Draft4Validator)
